/* Dump results to a file 
 * according to structure definitions in quartzcommon.h 
 */ 

#include <stdio.h>
#include "thread.h"
#include "synch.h"
#include <usclkc.h>
#include "quartzcommon.h"
#include "profile.h"
#include "internal.h"
#include "util.h"

static SynchProfile *synchList = NULL;
static FILE *ofd = NULL;
static int firstKid;

static int WasHit (p)
	ConcurrentData *p;
	{
	int i;

	for (i = 0; i < MaxProfilers; i++)
		if (p->hit[i] > 0)
			return(TRUE);
	return(FALSE);
	}

static void DumpBusy (p)
	BusyData *p;
	{
	int i;

	fprintf(ofd, "{{ /* BusyData */");
	for (i = 0; i < numProcessors; i++)
		fprintf(ofd, "%f, ", p->byEffP[i]);
	fprintf(ofd, "}}, \n");
	}

static void DumpData (p, ql)
	ConcurrentData *p;
	QueueLength *ql;
	{
	int i, j, k;

	fprintf(ofd, "{ /* MeasuredData */ \n");

	DumpBusy(&p->busy);

	fprintf(ofd, "{{ /* NominalData */ ");
	for (i = 0; i < NumWho; i++)
		{
		fprintf(ofd, "{ ");
		for (j = 0; j < NumStates; j++)
			{
			fprintf(ofd, "{ ");
			for (k = 0; k < 2; k++)
				fprintf(ofd, "/* %d %d %d */ %f, ", 
					i, j, k, p->nom.byNomP[i][j][k]);
			fprintf(ofd, "}, \n ");
			}
		fprintf(ofd, "}, \n");
		}
	fprintf(ofd, "}}, \n");

	fprintf(ofd, "{{ /* QueueLength */"); 
	if (ql == NULL)
		fprintf(ofd, "0");
	else
		for (i = 0; i < NumNumbers; i++)
			{
			fprintf(ofd, "{\n"); 
			for (j = 0; j < MaxNominalParallelism; j++)
				fprintf(ofd, "%f, ", ql->length[i][j]);
			fprintf(ofd, "},\n");
			}
	fprintf(ofd, "}}\n"); 

	fprintf(ofd, "},\n");
	}

static void DumpProcs ()
	{
	int i;

	fprintf(ofd, "static ProcData procData[] = {\n");
	for (i = 0; i < mNumProcIds; i++)
		if (WasHit(&procData[i]))
			{
			fprintf(ofd, "{ /* id */ %d, \n ", i);
			DumpData(&procData[i], (QueueLength *)NULL);
			fprintf(ofd, "},\n");
			}
	fprintf(ofd, "};\n\n");
	}

static void DumpObj (p)
	SynchProfile *p;
	{
	char *s;

	switch (p->kind)
		{
		case VarSpinLock:
			s = "VarSpinLock";
			break;
		case VarSpinDependency:
			s = "VarSpinDependency";
			break;
		case VarBlockLock:
			s = "VarBlockLock";
			break;
		case VarBlockDependency:
			s = "VarBlockDependency";
			break;
		case VarThread:
			s = "VarThread";
			break;
		}
	fprintf(ofd, "{\"%s\", %s, %d, %d,", p->name, s, 
			p->synchId, Id2NewId(p->parentID));

	if (p->samples)
		{
		fprintf(ofd, " TRUE, \n"); 
		DumpData(&p->samples->data, &p->samples->queue);
		}
	else
		fprintf(ofd, "FALSE, {0}");

	fprintf(ofd, "},"); 
	}

static void DumpObjs ()
	{
	SynchProfile *p, *last;
	int i, nsl;

	for (i = 0, last = synchList = NULL, nsl = 0; i < numProcessors; i++)
		for (p = processorList[i].synchList; p; p = p->next)
			{	/* put them in the list in the order of their ids */ 
			p->synchId = mNumProcIds + nsl++;
			if (last == NULL)
				last = synchList = p;
			else
				{
				last->dumpNext = p;
				last = p;
				}
			p->dumpNext = NULL;
			}

	fprintf(ofd, "static SynchData synchData[] = {\n");
	for (p = synchList; p; p = p->dumpNext)
		DumpObj(p);
	fprintf(ofd, "};\n\n");
	}

static void DumpKidList (k, caller)
	ChildData *k;
	int caller;
	{
	for (; k; k = k->next)
		{
		ASSERT(k->type == ChildDataType);
		if (k->calleeID != NoID)
			{
			if (!firstKid)
				{
				firstKid = TRUE;
				fprintf(ofd, " = {\n");
				}
			fprintf(ofd, "{ /*caller*/ %d, /* callee */ %d,\n",
				caller, Id2NewId(k->calleeID));
			DumpBusy(&k->busy);
			fprintf(ofd, "},\n");
			}
		}
	}

static void DumpKids ()
	{
	int i;
	SynchProfile *p;

	firstKid = FALSE;
	fprintf(ofd, "static KidData kidData[]");

	for (i = 0; i < mNumProcIds; i++)
		DumpKidList(&procData[i].kid, i);

	for (p = synchList; p; p = p->dumpNext)
		if (p->samples)
			DumpKidList(&p->samples->data.kid, p->synchId);
	
	if (firstKid)
		fprintf(ofd, "}");
	fprintf(ofd, ";\n\n");
	}

static void DumpGraphList (p)
	GraphEntry *p;
	{
	for (; p; p = p->next)
		if (p->num && (p->calleeID != NoID) && (p->callerID != NoID))
			fprintf(ofd, "{%d, %d, %d},\n", 
				Id2NewId(p->calleeID), Id2NewId(p->callerID), p->num);
	}

static void DumpTriples ()
	{
	int i;
	SynchProfile *p;

	fprintf(ofd, "static Triple triples[] = {\n");
	fprintf(ofd, "{%d, %d, %d},\n", IdleID, StartID, numProcessors);
	fprintf(ofd, "{%d, %d, %d},\n", ForkID, StartID, numProcessors);

	for (i = 0; i < pcTableSize; i++)
		DumpGraphList(&pcTable[i]);

	for (p = synchList; p; p = p->dumpNext)
		DumpGraphList(&p->g);
	
	fprintf(ofd, "};\n\n");
	}

void DumpInfo ()
	{
	ofd = MyOpen("quartzData.c", "w");

	fprintf(ofd, "#include \"quartzcommon.h\"\n\n");

	DumpProcs();

	DumpObjs();

	DumpKids();

	DumpTriples();

	fprintf(ofd, "QuartzData qdata = {%d, %d, %f, %d,\n", 
		numProcessors, mNumProcIds, timeDiff, numSamples);
	fprintf(ofd, "procData, sizeof(procData)/sizeof(ProcData),\n");
	fprintf(ofd, "synchData, sizeof(synchData)/sizeof(SynchData),\n");
	fprintf(ofd, "kidData, sizeof(kidData)/sizeof(KidData),\n");
	fprintf(ofd, "triples, sizeof(triples)/sizeof(Triple)};\n\n");

	MyClose(ofd);
	}

int Id2NewId (id)
	int id;
	{
	id &= AllOffMask;
	if (isSynchID(id))
		return(((SynchProfile *)id)->synchId);
	return(id);
	}
