/* Synchronization routines: for profiling, don't compile with -pg.
 * This would be a lot easier to express with objects --
 * SpinCondition, Condition, and Lock are fairly similar.
 *
 * The profiling assumes Mesa semantics -- to simplify managing the
 * stack of synchronization objects.  You have to do extra futzing
 * to get Hoare semantics.
 */

#include "thread.h"
#include "synch.h"
#include <usclkc.h>
#include "quartzcommon.h"
#include "profile.h"
#include "internal.h"

void SpinLockInit (s, name)
	register SpinLock *s;
	char *name;
	{
	s->type = SpinLockType;
	s->l = 0;
#ifdef PROFILE
	s->p = SynchInit(name, VarSpinLock);
#endif
	}

#ifdef PROFILE
void SpinLockAcquire (s)
	register SpinLock *s;
	{
	ASSERT(s->type == SpinLockType);
	PushOnIdStack((s)->p, (unsigned int)SpinState);
	AtomicIncrP(&((s)->p->number[0]));
	NowIdle();
	MY_LOCK(s);
	ASSERT(s->l == 1);
	SetStateBusy();
	NowBusy();
	}

void SpinLockRelease (s)
	register SpinLock *s;
	{
	ASSERT(s->type == SpinLockType);
	ASSERT(s->l == 1);
	MY_UNLOCK(s);
	PopOffIdStack();
	AtomicDecrP(&((s)->p->number[0]));
	}

void SpinLockRelBlock (s)
	register SpinLock *s;
	{
	ASSERT(s->type == SpinLockType);
	ASSERT(s->l == 1);
	MY_UNLOCK(s);
	PopOffIdStack();
	AtomicDecrP(&((s)->p->number[0]));
	SetStateBlocked();
	}
#endif

void SpinLockDispose (s)
	SpinLock *s;
	{
	ASSERT(s->type == SpinLockType);
#ifdef PROFILE
	SynchDispose(s->p);
#endif
	}

#ifndef PRESTO

void ConditionInit (c, s)
	Condition *c;
	char *s;
	{
	c->type = ConditionType;
	c->waiting = NULL;
#ifdef PROFILE
	c->p = SynchInit(s, VarBlockDependency);
#endif
	}

void ConditionDispose (c)
	Condition *c;
	{
	ASSERT(c->type == ConditionType);
#ifdef PROFILE
	SynchDispose(c->p);
#endif
	}

void ConditionSignal (c)
	Condition *c;
	{
	register Thread *t;

	ASSERT(c->type == ConditionType);
	t = c->waiting;
	if (t) 
		{
		ASSERT(t->type == ThreadType);
#ifdef PROFILE
		c->p->number[0]--;
#endif
		c->waiting = t->next;
		ReadyQPut(t);
		}
	}

void ConditionBroadcast (c)
	Condition *c;
	{
	register Thread *t;

	ASSERT(c->type == ConditionType);
#ifdef PROFILE
	c->p->number[0] = 0;
#endif
	for (t = c->waiting; t; t = c->waiting) 
		{
		ASSERT(t->type == ThreadType);
		c->waiting = t->next;
		ReadyQPut(t);
		}
	}

void ConditionWait (c, l)
	Condition *c;
	Lock *l;
	{
	register Thread *t = pP.thread;

	ASSERT(c->type == ConditionType);
#ifdef PROFILE
	c->p->number[0]++;
	PopOffIdStack();
	PushOnIdStack(c->p, BusyState);	
#endif
	ThreadPut(t, &c->waiting);
	SpinLockAcquire(&l->slock);
	if (t = l->waiting)
		{
#ifdef PROFILE
		l->p->number[0]--;
#endif
		l->waiting = t->next;
		ReadyQPut(t);
		}
	l->lock = 0;
	Block(&l->slock);
#ifdef PROFILE
	PopOffIdStack();	
#endif
	}

void SpinConditionWait (c, l)
	SpinCondition *c;
	SpinLock *l;
	{
	register Thread *t = pP.thread;

	ASSERT(c->type == ConditionType);
#ifdef PROFILE
	c->p->number[0]++;
	CallAndReplaceOnIdStack(c->p, BusyState);
#endif
	ThreadPut(t, &c->waiting);
	BlockNoProf(l);		/* spin lock is no longer on stack */
#ifdef PROFILE
	PopOffIdStack();
#endif
	}

/* A blocking lock, implemented Mesa-style */
void LockInit (l, s)
	Lock *l;
	char *s;
	{
	l->type = LockType;
	l->lock = 0;
	SpinLockInit(&l->slock, s);
	l->waiting = NULL;
#ifdef PROFILE
	l->p = SynchInit(s, VarBlockLock);
#endif
	}

void LockDispose (l)
	Lock *l;
	{
	ASSERT(l->type == LockType);
	SpinLockDispose(&l->slock);
#ifdef PROFILE
	SynchDispose(l->p);
#endif
	}

/* HACK! HACK! -- separate this from LockAcquire to fool the optimizer
 * into doing what it should do in the first place!
 */
void doit (t, l)
	Thread *t;
	Lock *l;
	{
#ifdef PROFILE
	l->p->number[0]++;
#endif
	ThreadPut(t, &l->waiting);
	Block(&l->slock);
	}

void LockAcquire (l)	
	Lock *l;
	{
	register Thread *t = pP.thread;

	ASSERT(l->type == LockType);
#ifdef PROFILE
	PushOnIdStack(l->p, BusyState);
#endif
	SpinLockAcquire(&l->slock);
	while (l->lock) 
		{
		doit(t, l);
		SpinLockAcquire(&l->slock);
		}
	l->lock = 1;
	SpinLockRelease(&l->slock);
#ifdef PROFILE
	PopOffIdStack();
#endif
	}

/* Pass the lock, but not the spinlock, if someone is waiting */
void LockRelease (l)
	Lock *l;
	{
	register Thread *t;

	ASSERT(l->type == LockType);
	ASSERT(l->lock == 1);
	SpinLockAcquire(&l->slock);
	if (t = l->waiting)
		{
#ifdef PROFILE
		l->p->number[0]--;
#endif
		l->waiting = t->next;
		ReadyQPut(t);
		}
	l->lock = 0;
	SpinLockRelease(&l->slock);
#ifdef PROFILE
	PopOffIdStack();
#endif
	}

static void BarInit (b, initialCount, name)
	Barrier *b;
	int initialCount;
	char *name;
	{
	b->type = BarrierType;
	b->count = b->initial = initialCount;
	b->iteration = 0;
	SpinLockInit(&b->l, name);
	}

void SpinBarrierInit (b, initialCount, name)
	SpinBarrier *b;
	int initialCount;
	char *name;
	{
	BarInit((Barrier *)b, initialCount, name);
	b->c.waiting = NULL;  /* needed for test in BarrierCheckIn */
#ifdef PROFILE
	b->c.p = SynchInit(name, VarSpinDependency);
#endif
	}

void BarrierInit (b, initialCount, name)
	Barrier *b;
	int initialCount;
	char *name;
	{
	BarInit(b, initialCount, name);
	SpinConditionInit(&b->c, name);
	}

void BarrierDispose (b)
	Barrier *b;
	{
	ASSERT(b->type == BarrierType);
	SpinLockDispose(&b->l);
	SynchDispose(b->c.p);
	}

int BarrierCheckIn (b)
	Barrier *b;
	{
	int v;

	ASSERT(b->type == BarrierType);
	v = b->iteration;
	SpinLockAcquire(&b->l);
#ifdef PROFILE
	b->c.p->number[1]++;
#endif
	if (--b->count == 0)
		{
		b->count = b->initial;
		b->iteration++;
		if (b->c.waiting)
			SpinConditionBroadcast(&b->c);
#ifdef PROFILE
		b->c.p->number[1] = 0;
#endif
		}
	SpinLockRelease(&b->l);
	return(v);
	}

void SpinBarrierWait (b, i)
	SpinBarrier *b;
	int i;
	{
	ASSERT(b->type == BarrierType);
#ifdef PROFILE
	PushOnIdStack(b->c.p, (unsigned int)SpinState);
	AtomicIncrP(&(b->c.p->number[0]));
	NowIdle();
#endif
	while (b->iteration == i)
		;
#ifdef PROFILE
	NowBusy();
	AtomicDecrP(&(b->c.p->number[0]));
	PopOffIdStack();
#endif
	}

void BarrierWait (b, i)
	Barrier *b;
	int i;
	{
	ASSERT(b->type == BarrierType);
#ifdef PROFILE
	PushOnIdStack(b->c.p, BusyState);
#endif
	if (b->iteration == i)
		{
		SpinLockAcquire(&b->l);
		if (b->iteration == i)
			SpinConditionWait(&b->c, &b->l); 	/* releases the lock */
		else
			SpinLockRelease(&b->l);
		}
#ifdef PROFILE
	PopOffIdStack();
#endif
	}

#endif /* ifndef PRESTO */
